/* **********************************************************
 * Copyright 1998 VMware, Inc.  All rights reserved. -- VMware Confidential
 * **********************************************************/

extern "C" {
/*
 * task.c --
 *
 *      Task initialization and switching routines between the host
 *      and the monitor. 
 *     
 *      A task switch: 
 *          -saves the EFLAGS,CR0,CR2,CR4, and IDT
 *          -jumps to code on the shared page
 *              which saves the registers, GDT and CR3
 *              which then restores the registers, GDT and CR3
 *          -restores the IDT,CR0,CR2,CR4 and EFLAGS
 *                
 *      This file is pretty much independent of the host OS.
 *      
 */

/* On Linux, must come before any inclusion of asm/page.h --hpreg */
#include "hostKernel.h"
#ifdef linux
#   include <linux/string.h>
#   include <linux/kernel.h>

#   ifdef USE_PERFCTRS_HOSTED
#      include "perfctr.h"
#   endif

#   define EXPORT_SYMTAB
#endif

#include "vmware.h"
#include "modulecall.h"
#include "vmx86.h"
#include "task.h"
#include "vm_asm.h"
#include "cpuid.h"
#include "hostif.h"
}

typedef struct SysenterStateV45 {
   uint32       esp;
   uint32       eip;
   Selector     cs;
   Bool         validEIP;
};

#if defined(_WIN64)
#include "vmmon-asm-x86-64.h"
#endif


/*
 * hostSysenter: values of host sysenter MSRs for each host cpu.
 */
static SysenterRegs *hostSysenter = 0;
static Bool hostMayUseSysenter = FALSE;
SysenterState VMCrossPageV45::vmmSysenter = { { 0 } };

/*
 *----------------------------------------------------------------------
 * TaskProbeSysenterMSR --
 *
 *    This function computes sysenterMSR. Note that we assume that if a
 *    host is ever going to use the sysenter MSRs, they will have been
 *    initialized here; if some future host lazily sets up the sysenter
 *    MSRs, this code will cause trouble.
 *
 *----------------------------------------------------------------------
 */
static void
TaskProbeSysenterMSR(void)
{
   CPUOem cpuOem = CPUID_GetOEM();
   uint32 features = CPUID_GetFeatures();
   uint32 version = CPUID_GetVersion();

   /*
    * Some old intel CPUs set the SEP bit, even though they are not
    * capable of sysenter. In particular, CPUs with family of 6, but
    * with the whole version string < 0x633 (family, model, stepping).
    * This check is also performed in vmx/vmcore/cpuid.c.
    */
   if (cpuOem == CPUID_OEM_INTEL && CPUID_FAMILY(version) == 6 &&
       (CPUID_MODEL(version) < 3 ||
        (CPUID_MODEL(version) == 3 && CPUID_STEPPING(version) < 3))) {

      Warning("VMMON: CPU too old to use sysenter (ok).\n");
   } else if (features & CPUID_FEATURE_COMMON_ID1EDX_SEP) {
      /* The CPU appears to support sysenter. */
      hostMayUseSysenter = 1;
   }
}


/*
 *------------------------------------------------------------------------------
 *
 *  Task_SysenterMSRInitialize --
 *
 *    Initialize the per-physical-cpu sysenter msr structures.
 *
 * Results:
 *    TRUE  -> Successfully initialized.
 *    FALSE -> Unable to allocate memory for sysenter msrs;
 *             system cannot continue to load.
 *
 * Side effects:
 *    None
 *  
 *------------------------------------------------------------------------------ */

Bool
Task_SysenterMSRInitialize(void)
{
   const unsigned cpus  = HostIF_NumOnlineLogicalCPUs();
   const unsigned bytes = cpus * sizeof(SysenterRegs);
   unsigned       i;
   
   TaskProbeSysenterMSR();
 
   /* Some linux kernels panic when allocating > 128Kb: ensure an
    * attempt is not made to run with ~1365 host cpus */
   ASSERT(bytes <= 131072);

   hostSysenter = (SysenterRegs *)HostIF_AllocKernelMem(bytes, TRUE);
   if (hostSysenter != NULL) {
      for (i = 0; i < cpus; ++i) {
         hostSysenter[i].sysenterCS  = SYSENTER_SENTINEL_CS;
         hostSysenter[i].sysenterEIP = SYSENTER_SENTINEL_EIP;
         hostSysenter[i].sysenterESP = SYSENTER_SENTINEL_ESP;
      }
      return TRUE;
   } else {
      Warning("sysenter msr initialization failure; "
              "unable to allocate space for values\n");
      return FALSE;
   }
}


/*
 *------------------------------------------------------------------------------
 *
 * TaskInitGuestSysenter  --
 *
 *    Initialize the crosspage sysenter msr state
 *
 * Results:
 *    none
 *
 * Side effects:
 *    None
 *  
 *------------------------------------------------------------------------------
 */

static INLINE_SINGLE_CALLER void
TaskInitGuestSysenter(SysenterState *sysenter)
{
   sysenter->hw.sysenterCS    = 0;
   sysenter->hw.sysenterEIP   = 0;
   sysenter->hw.sysenterESP   = 0;
   sysenter->requestedCS      = 0;
   sysenter->sysenterCacheMSR = TRUE;
}


/*
 *------------------------------------------------------------------------------
 *
 * Task_InitCrosspage --
 *
 *    Initialize the crosspage used to switch to the monitor task. 
 *
 * Results:
 *    0 on success
 *    != 0 on failure
 *
 * Side effects:
 *    None
 *  
 *------------------------------------------------------------------------------
 */

template <class VMCP> static int 
Task_InitCrosspage_TOT(VMDriver *vm,          // IN
                       InitBlock *initParams, // IN: Initial params from the VM 
		       int ws451)
{
   Vcpuid vcpuid;

   ASSERT(sizeof(VMCrossPageTOT) < PAGE_SIZE);
   ASSERT(MODULECALL_CROSS_PAGE_LEN == 1);
   for (vcpuid = 0; vcpuid < initParams->numVCPUs;  vcpuid++) {
      uintptr_t    cr3reg;
      LA           linearAddr;
      uint16       trReg;
      void        *iparm = (void *)(uintptr_t)initParams->crosspage[vcpuid];
      VMCP        *p = (VMCP*)HostIF_MapCrossPage(vm, iparm, &linearAddr);

      if (p == NULL) {
         return 1;
      }

      vm->crosspage[vcpuid] = p;
      GET_CR3(cr3reg);
      GET_TR(trReg);
#if defined(VM_X86_64)
      {
         MPN crossPageMPN = HostIF_LookupUserMPN(vm, (void *)(uintptr_t)initParams->crosspage[vcpuid]);
         int c32 = ws451 ? 1 : 2;
	 int c64 = ws451 ? 2 : 1;
      
         if ((int64)crossPageMPN <= 0) {
            return 1;
         }

         p->hostContext64.context.cr3 = cr3reg;
         p->hostVA64                  = (uintptr_t)p;

         /*
          * In order to make the transition to legacy mode, we must first turn
          * paging off on the processor. For this we need to be executing
          * code on an identity mapped page (the crosspage).
          */
         p->crosspageLinearAddr = (LA32)MPN_2_MA(crossPageMPN);

         /* Make sure there is no truncation */
         ASSERT((VA)p->crosspageLinearAddr == MPN_2_MA(crossPageMPN));      
         p->hostContext64VA = (uintptr_t)&p->hostContext64;

         /* XXX hack the initial monitor idt */
         p->monContext.idtr.dtr.offset = 0;
         p->monContext.idtr.dtr.limit = sizeof(Gate)*0x100 - 1;

         p->hostContext64.switchGdtr64.offset = p->crosspageLinearAddr + 
            offsetof(VMCP, switchGdt);
         p->hostContext64.switchGdtr64.limit = sizeof(Descriptor) * 4 - 1;

         /*
          * Set up data for worldswitch jumping between modes.  The
          * data is used in the dynamically loaded worldswitch code.
          * The offsets are set in
          * $VMTREE/vmcore/vmx/main/monitorHosted32.c:PatchVmm64Info().
          */
         p->hostContext64.jump64Code.selector = MAKE_SELECTOR(c64, SELECTOR_GDT, 0);
         p->hostContext64.jump32Code.selector = MAKE_SELECTOR(c32, SELECTOR_GDT, 0);
         p->hostContext64.switchDS            = MAKE_SELECTOR(3, SELECTOR_GDT, 0);
         p->hostContext64.tr                  = trReg;
         
         // null descriptor
         p->switchGdt[0].word[0] = 0;
         p->switchGdt[0].word[1] = 0;

         // 64 bit CS descriptor
         Desc_SetDescriptor(&p->switchGdt[c64].desc, 0, (VA32)0xfffff, CODE_DESC, 
                            1, 0, 1, 0, 1);
         // Set the L bit to indicate running 64 bit mode.
         p->switchGdt[c64].desc.longmode = 1;

         // 32 bit compatibility mode CS descriptor
         Desc_SetDescriptor(&p->switchGdt[c32].desc, 0, (VA32)0xfffff, CODE_DESC, 
                            1, 0, 1, 1, 1);
         
         // 32 bit compatibility mode DS descriptor
         Desc_SetDescriptor(&p->switchGdt[3].desc, 0, (VA32)0xfffff, DATA_DESC, 
                            1, 0, 1, 1, 1);
      }
#else
      {
         DTR32 idtrReg;

         GET_IDT(idtrReg);
         p->hostContext.task.cr3 = cr3reg;
         p->crosspageLinearAddr  = linearAddr;
         p->hostContextVA        = (uintptr_t)&p->hostContext;
         p->monContext.idtr.dtr  = idtrReg; /* irrelevant */
         p->hostContext.tr       = trReg;
      }
#endif

      p->irqRelocateOffset[0]  = IRQ_HOST_INTR1_BASE; 
      p->irqRelocateOffset[1]  = IRQ_HOST_INTR2_BASE;
      p->userCallRequest       = MODULECALL_USERCALL_NONE;
      p->moduleCallInterrupted = FALSE;
      p->tscAdjustment         = 0;
      TaskInitGuestSysenter(&p->vmmSysenter);
   }
   return 0;
}


#ifndef VM_X86_64
/*
 *------------------------------------------------------------------------------
 *
 * Task_InitCrosspage_V4
 *
 *    Initialize the crosspage used to switch to the monitor task. 
 *
 * Results:
 *    0 on success
 *    != 0 on failure
 *
 * Side effects:
 *    None
 *  
 *------------------------------------------------------------------------------
 */

static int 
Task_InitCrosspage_V4(VMDriver *vm,          // IN
                      InitBlock *initParams) // IN: Initial params from the VM 
{
   LA baseLinearAddr = HOST_KERNEL_VA_2_LA(0);
   uint32 cr3reg;
   Vcpuid vcpuid;

   ASSERT(sizeof (VMCrossPageV4) < PAGE_SIZE);
   ASSERT(MODULECALL_CROSS_PAGE_LEN==1);
   for (vcpuid = 0; vcpuid < initParams->numVCPUs;  vcpuid++) {
      /*
       * The physical memory associated to the crosspage has already been
       * locked in MonitorAllocateMemory() --hpreg
       */
      VMCrossPageV4 *p = (VMCrossPageV4*)HostIF_UserToDriverPtr(vm, initParams->crosspage[vcpuid]);
      if (p == NULL) {
         return 1;
      }
      vm->crosspage[vcpuid] = p;
      GET_CR3(cr3reg);
      p->hostContext.task.cr3 = cr3reg;
      p->hostVA = (VA32)p;
      p->hostContextVA = (uintptr_t)&p->hostContext;
      p->crosspageLinearAddr = baseLinearAddr + (uintptr_t)p;
      {
         DTR32 idtrReg;
         GET_IDT(idtrReg);
         p->monContext.idtr.dtr = idtrReg; /* irrelevant */
      }

      {
         uint16 trReg;
         GET_TR(trReg);
         p->hostContext.tr = trReg;
      }

      p->irqRelocateOffset[0] = IRQ_HOST_INTR1_BASE; 
      p->irqRelocateOffset[1] = IRQ_HOST_INTR2_BASE;

      p->userCallRequest = MODULECALL_USERCALL_NONE;
      p->moduleCallInterrupted = FALSE;
    }
   return 0;
}
#endif

static Bool
DisableNMIDelivery(volatile uint32 *regPtr) // IN/OUT
{
   uint32 reg;
   
   reg = *regPtr;
   if ((APIC_LVT_DELVMODE(reg) == APIC_LVT_DELVMODE_NMI) &&
       (! APIC_LVT_ISMASKED(reg))) {
      *regPtr = reg | APIC_LVT_MASK;
      return TRUE;
   }
   return FALSE;
}

#define NMI_LINT0	(0x01)
#define NMI_LINT1	(0x02)
#define NMI_TIMER	(0x04)
#define NMI_ERR		(0x08)
#define NMI_PC		(0x10)
#define NMI_THERM	(0x20)

static unsigned int
DisableNMI2(VMDriver *vm)   // IN
{
   unsigned int nmi = 0;
   if (vm->hostAPIC) {
      switch (APIC_MAX_LVT(vm->hostAPIC)) {
         default:
         case 5:
	    if (DisableNMIDelivery(&APIC_THERM_REG(vm->hostAPIC))) {
	       nmi |= NMI_THERM;
	    }
	 case 4:
	    if (DisableNMIDelivery(&APIC_PC_REG(vm->hostAPIC))) {
	       nmi |= NMI_PC;
	    }
	 case 3:
	 case 2:
	 case 1:
	 case 0:
	    if (DisableNMIDelivery(&APIC_LINT0_REG(vm->hostAPIC))) {
	       nmi |= NMI_LINT0;
	    }
	    if (DisableNMIDelivery(&APIC_LINT1_REG(vm->hostAPIC))) {
	       nmi |= NMI_LINT1;
	    }
	    if (DisableNMIDelivery(&APIC_TIMER_REG(vm->hostAPIC))) {
	       nmi |= NMI_TIMER;
	    }
	    if (DisableNMIDelivery(&APIC_ERR_REG(vm->hostAPIC))) {
	       nmi |= NMI_ERR;
	    }
	    break;
      }
   }
   return nmi;
}

static unsigned int
DisableNMI(VMDriver *vm)   // IN
{
   unsigned int nmi = DisableNMI2(vm);
   unsigned int nmi2;
   
   /* Notify about race between timer and NMI on P4. It disables PC interrupt
      delivery when timer triggers, but actual delivery of NMI can race with us,
      and NMI handler can enable it under us. So we do it twice, hoping for
      the best... */
   nmi2 = DisableNMI2(vm);
   return nmi | nmi2;
}


static void
RestoreNMI(VMDriver *vm,  // IN
           unsigned int nmi) // IN
{
#define RestoreNMIDelivery(cond, apicr)				\
   do {								\
      if (nmi & cond) {						\
         uint32 reg;						\
	 							\
         reg = apicr;						\
         apicr = reg & ~APIC_LVT_MASK;				\
      }								\
   } while (0)

   RestoreNMIDelivery(NMI_LINT0, APIC_LINT0_REG(vm->hostAPIC));
   RestoreNMIDelivery(NMI_LINT1, APIC_LINT1_REG(vm->hostAPIC));
   RestoreNMIDelivery(NMI_TIMER, APIC_TIMER_REG(vm->hostAPIC));
   RestoreNMIDelivery(NMI_ERR,   APIC_ERR_REG(vm->hostAPIC));
   RestoreNMIDelivery(NMI_PC,    APIC_PC_REG(vm->hostAPIC));
   RestoreNMIDelivery(NMI_THERM, APIC_THERM_REG(vm->hostAPIC));
#undef RestoreNMIDelivery
}


void
Task_SysenterMSRReset(void)
{
   if (hostSysenter != NULL) {
      HostIF_SysenterReset(HostIF_NumOnlineLogicalCPUs(), hostSysenter);
      HostIF_FreeKernelMem(hostSysenter);
   }
}


#define SYSENTER_LOG_MSR(_f, _val, _msr)                                \
{                                                                       \
   if (0) {                                                             \
      const char          *f         = _f;                              \
      const uint32         val       = _val;                            \
      const char          *msr       = _msr;                            \
      const unsigned int   threshold = 100000;                          \
      static unsigned int  c         = 0;                               \
      if ((++c % threshold) == 0U) {                                    \
         Warning("%32s[%8d]: wrote '%#x' into sysenter MSR '%s'\n",     \
                 f, c, val, msr);                                       \
      }                                                                 \
   }                                                                    \
}

#define SYSENTER_SET_MSR(_msr, _oldRegs, _newRegs)              \
{                                                               \
   const uint32 oldReg = (_oldRegs)->sysenter##_msr;            \
   const uint32 newReg = (_newRegs)->sysenter##_msr;            \
   if (oldReg != newReg) {                                      \
      HostIF_WRMSR(MSR_SYSENTER_##_msr, newReg);                \
      SYSENTER_LOG_MSR("set eip/esp", newReg, #_msr);           \
   }                                                            \
}

#define SYSENTER_SET_MSR_CS(_oldCS, _newCS)                     \
{                                                               \
   const Selector oldCS = _oldCS;                               \
   const Selector newCS = _newCS;                               \
   if (oldCS != newCS) {                                        \
      HostIF_WRMSR(MSR_SYSENTER_CS, newCS);                     \
      SYSENTER_LOG_MSR("set cs", newCS, "CS");                  \
   }                                                            \
}


/*
 *-------------------------------------------------------------- ---------------
 *
 * SysenterValidateVmmState  --
 *
 *     Ensure that the physical sysenter msrs contain the correct
 *     values as expected by the vmm.  If the values are not correct,
 *     log a warning and reset the value.
 *     
 *     For internal diagnostic use only.  Should normally be disabled.
 *
 *-----------------------------------------------------------------------------
 */

static INLINE_SINGLE_CALLER void
SysenterValidateVmmState(const SysenterState *vmmState)
{
#define SYSENTER_VALIDATE_MSR(_msr, _expected)                          \
{                                                                       \
   const uint32 expected = _expected;                                   \
   const uint32 val = (uint32)HostIF_RDMSR(MSR_SYSENTER_##_msr);        \
   if (val != expected) {                                               \
      Warning("sysenter MSR '%s' is %#x; expected %#x\n",               \
              #_msr, val, expected);                                    \
      HostIF_WRMSR(MSR_SYSENTER_##_msr, expected);                      \
   }                                                                    \
}
   if (vmx86_debug) {
      SYSENTER_VALIDATE_MSR(CS, vmmState->hw.sysenterCS);
      if (vmmState->hw.sysenterCS != 0) {
         SYSENTER_VALIDATE_MSR(EIP, vmmState->hw.sysenterEIP);
         SYSENTER_VALIDATE_MSR(ESP, vmmState->hw.sysenterESP);
      }
   }
#undef SYSENTER_VALIDATE_MSR
}


/*
 *-----------------------------------------------------------------------------
 *
 * SysenterSaveHostState  --
 *
 *     If required, save the host state before returning to the monitor.
 *     
 *     If not required, diagnostic code checks if the current hardware
 *     MSR values match the cached values.
 *
 *-----------------------------------------------------------------------------
 */

static INLINE_SINGLE_CALLER void
SysenterSaveHostState(SysenterRegs        *hostMSR,
                      const SysenterState *vmmState)
{
   if (UNLIKELY(!vmmState->sysenterCacheMSR ||
                (hostMSR->sysenterCS == SYSENTER_SENTINEL_CS))) {
      hostMSR->sysenterESP = (uint32)HostIF_RDMSR(MSR_SYSENTER_ESP);
      hostMSR->sysenterEIP = (uint32)HostIF_RDMSR(MSR_SYSENTER_EIP);
      hostMSR->sysenterCS  = (Selector)HostIF_RDMSR(MSR_SYSENTER_CS);
   }
}


/*
 *-----------------------------------------------------------------------------
 *
 * SysenterReturnToMonitor  --
 *
 *     Set up the sysenter MSR values for use by the vmm.
 *
 *-----------------------------------------------------------------------------
 */

static INLINE_SINGLE_CALLER void
SysenterReturnToMonitor(SysenterRegs  *hostMSR,
                        SysenterState *vmmState)
{
   SysenterSaveHostState(hostMSR, vmmState);
   if (vmmState->requestedCS != 0) {
      /* Guest using sysenter hardware; update %eip, %esp.
       *
       * It is tempting to believe that one need only write %eip/%esp
       * when the guest OS changes them and the host does not use the
       * sysenter hardware.  However, this is not the case: consider a
       * UP VM which uses sysenter running on a host which does not
       * use sysenter.  In this scenario, the guest OS can set the
       * %eip & %esp msrs on host cpu0, and then get migrated to host
       * cpu1 (where the msr values would not be correct).
       */
      SYSENTER_SET_MSR(ESP, hostMSR, &vmmState->hw);
      SYSENTER_SET_MSR(EIP, hostMSR, &vmmState->hw);
   }
   SYSENTER_SET_MSR_CS(hostMSR->sysenterCS, vmmState->requestedCS);

   /* Update the hardware state to reflect the requested CS value.
    * This can only be performed after the host sysenter state has
    * been saved (Consider a host which has %cs as 0, and the guest
    * sets %cs to some non-zero value, and then resets it back to
    * zero.  If the 'requested' value is used to compare against the
    * host value (rather than the value actually in the hardware),
    * then a mismatch will occur).
    */
   if (vmmState->hw.sysenterCS != vmmState->requestedCS) {
      vmmState->hw.sysenterCS  = vmmState->requestedCS;
   }
   SysenterValidateVmmState(vmmState);
}


/*
 *-----------------------------------------------------------------------------
 *
 * SysenterReturnFromMonitor  --
 * 
 *      Restore host sysenter state
 *
 *-----------------------------------------------------------------------------
 */

static INLINE_SINGLE_CALLER void
SysenterReturnFromMonitor(SysenterRegs        *hostMSR,
                          const SysenterState *vmmState)
{
   if (!vmmState->sysenterCacheMSR || hostMSR->sysenterCS != 0) {
      /* Host is using sysenter hardware, or caching disabled. */
      SYSENTER_SET_MSR(ESP, &vmmState->hw, hostMSR);
      SYSENTER_SET_MSR(EIP, &vmmState->hw, hostMSR);
   }
   SYSENTER_SET_MSR_CS(vmmState->hw.sysenterCS, hostMSR->sysenterCS);
}


/*
 *-----------------------------------------------------------------------------
 *
 * SysenterReturnToMonitor  --
 *
 *     This function saves the host OS' sysenter state prior to
 *     returning to the monitor.
 *
 *     The monitor does not know whether the host uses the MSRs. Thus,
 *     in order to avoid unnecessary writes to MSR_SYSENTER_CS, we
 *     clear this MSR here if the monitor does not use it but the host
 *     has set it to a non-zero value.
 *
 *-----------------------------------------------------------------------------
 */

static INLINE_SINGLE_CALLER void
SysenterReturnToMonitorOld(SysenterStateV45 *hostState,  // OUT
                           int monUsesSysenter)          // IN
{
   if (!hostMayUseSysenter) return;
   hostState->validEIP = 0;
   hostState->cs = (Selector)HostIF_RDMSR(MSR_SYSENTER_CS);
   if (hostState->cs != 0) {
      if (monUsesSysenter) {
         hostState->esp = (uint32)HostIF_RDMSR(MSR_SYSENTER_ESP);
         hostState->eip = (uint32)HostIF_RDMSR(MSR_SYSENTER_EIP);
	 hostState->validEIP = 1;
      } else {
         /* Monitor does not use sysenter, but host does.  Clear
          * MSR_SYSENTER_CS to induce #GP if sysenter used in the
          * vmm/guest.
          */
         HostIF_WRMSR(MSR_SYSENTER_CS, 0);
      }
   }
}


/*
 *-----------------------------------------------------------------------------
 *
 * SysenterReturnFromMonitor  --
 *
 *      Restore host sysenter state
 *
 *-----------------------------------------------------------------------------
 */

static INLINE_SINGLE_CALLER void
SysenterReturnFromMonitorOld(const SysenterStateV45 *hostState,  // IN
                             int monUsesSysenter)                // IN
{
   if (!hostMayUseSysenter) return;
   if (monUsesSysenter || hostState->cs != 0) {
      HostIF_WRMSR(MSR_SYSENTER_CS, hostState->cs);
   }
   if (monUsesSysenter && hostState->validEIP) {
      HostIF_WRMSR(MSR_SYSENTER_ESP, hostState->esp);
      HostIF_WRMSR(MSR_SYSENTER_EIP, hostState->eip);
   }
}


#ifdef VM_X86_64
/*
 * From an email with Petr regarding gcc's handling of the stdcall
 * attribute for x86-64:
 *
 *    As far as I can tell, for x86_64 there is only one calling
 *    convention:
 *       On Linux rdi/rsi/rdx/rcx/r8d/r9d for <= 6 arguments,
 *       others always on stack, caller always adjusts stack.
 *
 *       On Windows it is rcx/rdx/r8d/r9d for <= 4 arguments, rest on
 *       stack.  When more than 4 arguments are passed, spill space is
 *       reserved on the stack for the register arguments.  Argument
 *       5 is accessed at (5 * 8)(rsp).
 */

template <class VMCP>
static INLINE_SINGLE_CALLER void
SwitchToMonitor32(VMCP *crosspage, int ws451)
{
   VA32 monContextVA = VPN_2_VA(crosspage->crosspageStart()) + 
   		       offsetof(VMCP, monContext);
   PA monContextPA   = crosspage->crosspageLinearAddr + 
                       offsetof(VMCP, monContext);
#if defined(__GNUC__)
   if (ws451) {
   __asm__ __volatile__(
      "call  *%4"     "\n\t"
      :
      : "a" (&crosspage->hostContext64),
        "b" (monContextPA),
        "c" (monContextVA),
        "d" (crosspage->hostSwitchCR3),
        "g" (crosspage->contextSwitch.hostXToVmm32)
      : "memory", "cc"
   );
   } else {
   __asm__ __volatile__(
      "pushq  %4"     "\n\t"
      "call  *%5"     "\n\t"
      :
      : "a" (&crosspage->hostContext64),
        "b" (monContextPA),
        "c" (monContextVA),
        "d" (crosspage->hostSwitchCR3),
        "g" ((uint64)crosspage->hostContext64.context.cs),
        "r" (crosspage->contextSwitch.hostXToVmm32)
      : "memory", "cc"
   );
   }
#else
   SWITCH_TO_MONITOR((uint64)&crosspage->hostContext64,
                     monContextPA,
                     monContextVA,
                     crosspage->hostSwitchCR3,
                     (uint64)crosspage->hostContext64.context.cs,
                     (uint64)crosspage->contextSwitch.hostXToVmm32);
#endif
}


template <class VMCP>
static INLINE_SINGLE_CALLER void
SwitchToMonitor64(VMCP *crosspage)
{
#if defined(__GNUC__)
   __asm__ __volatile__(
      "pushq %2"       "\n\t"
      "call *%3"       "\n\t"
      :  
      : "a" (&crosspage->hostContext64),
        "b" (&crosspage->monContext64),
        "g" ((uint64) crosspage->hostContext64.context.cs),
        "r" ((uint8 *)crosspage->contextSwitch.vmm64ToHost64)
      : "memory", "cc"
   );
#else
#error No compiler defined for SwitchToMonitor64
#endif
   return;
}


static INLINE_SINGLE_CALLER void
SwitchToMonitor32(VMCrossPageTOT *crosspage, int dummy)
{
#if defined(__GNUC__)
#define PROTOPARMS32  uint32       zero0   /* rdi */,   \
                      uint32       zero1   /* rsi */,   \
                      uint32       zero2   /* rdx */,   \
                      VMCrossPageTOT *cp      /* rcx */
#define PARMS32 0 /* rdi */, 0 /* rsi */, 0 /* rdx */, crosspage /* rcx */
#elif defined(_MSC_VER)
#define PROTOPARMS32 VMCrossPageTOT *cp /* rcx */
#define PARMS32 crosspage /* rcx */
#else
#error No compiler defined for SwitchToMonitor32
#endif
   typedef void (*SwitchFn)(PROTOPARMS32);
   const uint8    *addr        = &crosspage->contextSwitch.hostXToVmm32[0];
   const SwitchFn  worldSwitch = (SwitchFn)addr;
   worldSwitch(PARMS32);
#undef PROTOPARMS32
#undef PARMS32
}


static INLINE_SINGLE_CALLER void
SwitchToMonitor64(VMCrossPageTOT *crosspage)
{
#if defined(__GNUC__)
   typedef void (*SwitchFn)(uint32   zero0  /* rdi */,
                            uint32   zero1  /* rsi */,
                            uint64   src    /* rdx */,
                            uint64   dest   /* rcx */,
                            Selector hostCS /* r8d */);
#define SWITCHFN64(_fn, _src, _dest, _hostCS)           \
               _fn(0       /* rdi */, 0     /* rsi */,  \
                   _src    /* rdx */, _dest /* rcx */,  \
                   _hostCS /* r8d */)
#elif defined(_MSC_VER)
   typedef void (__cdecl *SwitchFn)(uint64   dest   /* rcx */,
                                    uint64   src    /* rdx */,
                                    Selector hostCS /* r8d */);
#define SWITCHFN64(_fn, _src, _dest, _hostCS)           \
               _fn(_dest   /* rcx */, _src /* rdx */,   \
                   _hostCS /* r8d */)
#else
#error No compiler defined for SwitchToMonitor64
#endif

   const uint8    *worldSwitch = crosspage->contextSwitch.vmm64ToHost64;
   const SwitchFn  fn          = (SwitchFn)worldSwitch;
   const uint64    hostContext = (uint64)&crosspage->hostContext64;
   const uint64    monContext  = (uint64)&crosspage->monContext64;
   const Selector  hostCS      = crosspage->hostContext64.context.cs;

   SWITCHFN64(fn, hostContext, monContext, hostCS);
#undef SWITCHFN64
}


template <class VMCP>
static INLINE_SINGLE_CALLER void
SwitchToMonitor(VMCP *crosspage, int ws451)
{
   if (crosspage->runVmm64) {
      SwitchToMonitor64(crosspage);
   } else {
      SwitchToMonitor32(crosspage, ws451);
   }
}
#else

static INLINE_SINGLE_CALLER void
SwitchToMonitor32(unsigned char *codePtr,
                  ContextInfoV45 *src,
                  ContextInfoV45 *dst,
                  uint32 newDstVA)
{
#ifdef __GNUC__
   uint32 dummy1;
   uint32 dummy2;

   /* Checked against the Intel manual and GCC --hpreg */
   __asm__ __volatile__(
      "pushl %2"       "\n\t"
      "pushl %3"       "\n\t"
      "pushl %4"       "\n\t"
      "pushl %5"       "\n\t"
      "call *%6"       "\n\t"
      "popl %0"        "\n\t"
      "popl %1"        "\n\t"
      "addl %7, %%esp"
      /* 
       * Unfortunately there is no way to tell to gcc that we are 
       * using 'pushl' and so %esp-relative addressing does not
       * work. To prevent miscompilation we must limit ourselves
       * to registers and immediate values only.
       *
       * We do not have to use early-clobber, as %0 & %1 are used
       * only before %7, and %7 is immediate value.
       */
      : "=b" (dummy1),
        "=a" (dummy2)
      : "g" (newDstVA),
        "ri" (dst),
        "ri" (src),
        "ri" (0),
        "r" (codePtr),
        "i" ((1 + 1 + 3) * 4) // retval + segment + 3 arguments pushed
      : "memory", "cc"
   );
#elif _MSC_VER
   __asm push newDstVA
   __asm push dst
   __asm push src
   __asm push 0
   __asm call codePtr
   __asm pop ebx
   __asm pop eax
   __asm add esp, 20
#else
#   error No compiler defined for SwitchToMonitor
#endif
}

/* gcc 3.4.0 cannot handle always_inline in templates */
#if __GNUC__ == 3 && __GNUC_MINOR__ == 4
#define TEMPLATE_INLINE_SINGLE_CALLER /* I'm giving up... Please do not use gcc 3.4 if you want stable compiler */
#else
#define TEMPLATE_INLINE_SINGLE_CALLER INLINE_SINGLE_CALLER
#endif
template <class VMCP>
static TEMPLATE_INLINE_SINGLE_CALLER void
SwitchToMonitor(VMCP *crosspage, int dummy)
{
   SwitchToMonitor32(crosspage->contextSwitch.hostXToVmm32, &crosspage->hostContext, &crosspage->monContext, 
   		     VPN_2_VA(crosspage->crosspageStart()) + offsetof(VMCP, monContext));
}

/*
 *-----------------------------------------------------------------------------
 *
 * SwitchToMonitor --
 *
 *      Wrapper that calls code to switch from the host to the monitor.
 *
 * Side effects:
 *      None for the module
 *
 *-----------------------------------------------------------------------------
 */
static INLINE_SINGLE_CALLER void
SwitchToMonitor(VMCrossPageTOT *crosspage, int dummy)
{
#if defined(__GNUC__)
#define SWITCHFN (__attribute__((stdcall, regparm(0))) *SwitchFn)   
#elif defined(_MSC_VER)
#define SWITCHFN (__stdcall *SwitchFn)   
#else
#error No compiler defined for 32-bit SwitchToMonitor
#endif
   typedef void SWITCHFN(Selector     cs,
                         ContextInfo *src,
                         ContextInfo *dst,
                         uint32       newDstVA);

   ContextInfo    *src     = &(crosspage->hostContext);
   ContextInfo    *dst     = &(crosspage->monContext);
   const uint8    *codePtr = &crosspage->contextSwitch.hostXToVmm32[0];
   const Selector  hostCS  = crosspage->hostContext.task.cs;
    /*
    * newDstVA and vm->crosspage->monContext point to the same
    * location.  The latter refers to the monitor's virtual address
    * space.
    */
   uint32 newDstVA  = (uint32)(VPN_2_VA(crosspage->crosspageStart()) +
                               offsetof(VMCrossPageTOT, monContext));    
   SwitchFn fn = (SwitchFn)codePtr;
   fn(hostCS, src, dst, newDstVA);
#undef SWITCHFN
}

#endif // VM_X86_64

/*
 *-----------------------------------------------------------------------------
 *
 * Task_Switch --
 *
 *      Switches from the host context into the monitor
 *      context. Think of it as a coroutine switch that changes
 *      not only the registers, but also the address space
 *      and all the hardware state.
 *
 * Results:
 *      Next module call (or user call for that matter) is
 *      returned.
 *
 * Side effects:
 *      Jump to the other side. Has no direct effect on the
 *      host-visible state except that it might generate an interrupt.
 *
 *-----------------------------------------------------------------------------
 */

void 
Task_Switch_TOT(VMDriver *vm,  // IN
                Vcpuid vcpuid) // IN
{
   uintptr_t   flags, cr0reg, cr2reg, cr4reg, new_cr4;
   uintptr_t   drReg;
#ifdef VM_X86_64
   uint64      gs64, fs64;
#endif
   DTR32       gdtreg;
   Descriptor *desc;
   VA          vAddr;
   Selector    cs, gs, fs;
   uint16 trReg;   
   uint16 ldtreg;
   unsigned int lint;
   VMCrossPageTOT *crosspage = (VMCrossPageTOT*)vm->crosspage[vcpuid];
   SysenterRegs *hostMSR;

#if defined(linux) && defined(USE_PERFCTRS_HOSTED)
   PerfCtr_SnapshotFromHost();
#endif
   lint = DisableNMI(vm);
   SAVE_FLAGS(flags); 
   CLEAR_INTERRUPTS();

   vm->currentHostCpu[vcpuid] = HostIF_GetCurrentPCPU();
   ASSERT(hostSysenter != NULL);
   hostMSR = hostSysenter + vm->currentHostCpu[vcpuid];
   
   /*
    * Save CR state (not CR3!)
    */
   GET_CR0(cr0reg);
   GET_CR2(cr2reg);
   GET_CR4(cr4reg);

   /* Ensure global pages are flushed */
   new_cr4 = cr4reg & ~CR4_PGE;
   SET_CR4(new_cr4);

   crosspage->switchHostCR4 = new_cr4;

   /*
    * Save DR7 since we need to disable debug breakpoints during the
    * world switch code.  Save DR6 in order to accomodate the ICEBP
    * instruction.  All other debug registers are saved lazily by the
    * monitor prior to first use.  NOTE: Since monitor is in legacy
    * mode save all DRs for x86_64 here.
    */

   crosspage->hostDRSaved = 0;

#define SAVE_DR_SHADOW(n) \
                GET_DR##n(drReg); \
                crosspage->hostDR[n] = drReg; \
                crosspage->shadowDR[n] = (uint32) drReg; \
                crosspage->hostDRSaved |= (1 << n);

   if (vm_x86_64) {
      SAVE_DR_SHADOW(0);
      SAVE_DR_SHADOW(1);
      SAVE_DR_SHADOW(2);
      SAVE_DR_SHADOW(3);
   }

   SAVE_DR_SHADOW(6);
   SAVE_DR_SHADOW(7);
   
#undef SAVE_DR_SHADOW
   
   /*
    * Try to disable debug exceptions during the switch.
    * Unfortunately we cannot do this reliably, as the host
    * may have set DR7_GD.  This will cause the SET_DR7 to
    * trap, and the host trap handler can then put whatever
    * it wants in DR7 and resume after the SET_DR7.  We fix
    * this in the monitor, with our own trap handler.
    */

   if (UNLIKELY(crosspage->hostDR[7] & DR7_ENABLED)) {
      SET_DR7(DR7_DEFAULT);
      crosspage->shadowDR[7] = DR7_DEFAULT;
   }

#ifdef VM_X86_64
   GET_GDT(crosspage->hostContext64.gdtr64.dtr);
#else
   GET_GDT(gdtreg);
#endif
   GET_LDT(ldtreg); 

   /*
    * GS and FS are saved outside of the SwitchToMonitor() code to 1)
    * minimize the amount of code handled there and 2) prevent us from
    * faulting if they happen to be in the LDT (since the LDT is saved and
    * restored here too) and 3) make sure that if we do fault (because the
    * uncached descriptor for GS or FS's Selector has become invalid) we
    * fault in the host's context rather than the monitor or, worse,
    * never-never land. --Jeremy.
    */
#ifdef VM_X86_64
   gs64 = GET_GS64();
   fs64 = GET_FS64();
#endif
   gs = GET_GS();
   fs = GET_FS();
   
   SysenterReturnToMonitor(hostMSR, &crosspage->vmmSysenter);

   /*
    * Mark our task descriptor as unused, or we won't be
    * able to come back to it (SET_TR set the busy bit)
    */
   
   GET_TR(trReg);
   if (vm_x86_64) {
      crosspage->hostContext64.tr = trReg;
   } else {
      crosspage->hostContext.tr = trReg;
   }

   if (trReg) {
      if (vm_x86_64) {
         vAddr = HOST_KERNEL_LA_2_VA(crosspage->hostContext64.gdtr64.dtr.offset) + trReg;
      } else {
         vAddr = HOST_KERNEL_LA_2_VA(gdtreg.offset) + trReg;
      }
      desc = (Descriptor *)(vAddr);
      if (Desc_Type(desc) == TASK_DESC_BUSY) {
         Desc_SetType(desc, TASK_DESC);
      }
   }

   /*
    * Save the CS on this side
    */
   cs = GET_CS();
   if (vm_x86_64) {
      crosspage->hostContext64.context.cs = cs;
   } else {
      crosspage->hostContext.task.cs = cs;
   }

   /* See comment for SwitchToMonitor() */
   SwitchToMonitor(crosspage, 0);

   SysenterReturnFromMonitor(hostMSR, &crosspage->vmmSysenter);

   SET_CR0(cr0reg);
   SET_CR2(cr2reg);
   SET_CR4(cr4reg);
   /* The monitor shouldn't modify CR8 */
   
   SET_LDT(ldtreg);

   /* restore fs/gs must come before 64 bit fs/gs restore */
   SET_FS(fs);
   SET_GS(gs);
#ifdef VM_X86_64
   SET_FS64(fs64);
   SET_GS64(gs64);
#endif

#ifdef VM_X86_64
#define CAST_HOST_DR(x) (x)
#else
#define CAST_HOST_DR(x) ((uint32) x)
#endif

#define RESTORE_DR(n) if ((crosspage->hostDRSaved & (1 << n)) && \
                          (crosspage->hostDR[n] != crosspage->shadowDR[n])) { \
                          drReg = CAST_HOST_DR(crosspage->hostDR[n]); \
                          SET_DR##n(drReg); \
                      }

   RESTORE_DR(0);
   RESTORE_DR(1);
   RESTORE_DR(2);
   RESTORE_DR(3);
   RESTORE_DR(6);
   RESTORE_DR(7);

   if (UNLIKELY(crosspage->restoreGeneralDetect)) {
      crosspage->restoreGeneralDetect = 0;
      drReg = CAST_HOST_DR(crosspage->hostDR[7]) | DR7_GD;
      SET_DR7(drReg);
   }

#undef RESTORE_DR
#undef CAST_HOST_DR

   ASSERT_NO_INTERRUPTS();

   if (crosspage->moduleCallType == MODULECALL_TOT_INTR) {
      /*
       * Note we must do the RAISE_INTERRUPT before ever enabling
       * interrupts or bad things have happened (might want to know exactly
       * what bad things btw).
       * Note2: RAISE_INTERRUPT() only takes an constant and hence with switch
       * statement.
       */

#define IRQ_INT(_x) case _x: RAISE_INTERRUPT(_x); break 
#define IRQ_INT2(_x) IRQ_INT(_x); IRQ_INT(_x + 1)
#define IRQ_INT4(_x) IRQ_INT2(_x); IRQ_INT2(_x + 2)
#define IRQ_INT8(_x) IRQ_INT4(_x); IRQ_INT4(_x + 4)
#define IRQ_INT16(_x) IRQ_INT8(_x); IRQ_INT8(_x + 8)
#define IRQ_INT32(_x) IRQ_INT16(_x); IRQ_INT16(_x + 16)

      switch (crosspage->args[0]) {
	 // These are the general IO interrupts
	 // It would be nice to generate this dynamically, but see Note2 above.

	 /*
	  * Pass Machine Check Exception (Interrupt 0x12) to the host.
	  * See bug #45286 for details.
	  */
	 IRQ_INT(0x12);

         /*
          * pass the reserved vectors (20-31) as well. amd64 windows
          * generates these.
          */
	 IRQ_INT8(0x14);
	 IRQ_INT4(0x1c);

	 IRQ_INT32(0x20);
	 IRQ_INT32(0x40);
	 IRQ_INT32(0x60);
	 IRQ_INT32(0x80);
	 IRQ_INT32(0xa0);
	 IRQ_INT32(0xc0);
	 IRQ_INT32(0xe0);

      default: 
	 /*
	  * XXXX nt
	  * running on a 2 processor machine we hit this Panic with int 0xD1 0x61 ...
	  */
	 Warning("Received Unexpected Interrupt: 0x%X in Task_Switch()\n", crosspage->args[0]);
	 Panic("Received Unexpected Interrupt: 0x%X\n", crosspage->args[0]);
      }
   }
   
   vm->currentHostCpu[vcpuid] = INVALID_HOST_CPU;

   RESTORE_FLAGS(flags);
   RestoreNMI(vm, lint);
#if defined(linux) && defined(USE_PERFCTRS_HOSTED)
   PerfCtr_SnapshotFromMonitor();
#endif
}


/*
 *-----------------------------------------------------------------------------
 *
 * Task_Switch --
 *
 *      Switches from the host context into the monitor
 *      context. Think of it as a coroutine switch that changes
 *      not only the registers, but also the address space
 *      and all the hardware state.
 *
 * Results:
 *      Next module call (or user call for that matter) is
 *      returned.
 *
 * Side effects:
 *      Jump to the other side. Has no direct effect on the
 *      host-visible state except that it might generate an interrupt.
 *
 *-----------------------------------------------------------------------------
 */

void 
Task_Switch_V45(VMDriver *vm,  // IN
                Vcpuid vcpuid) // IN
{
   uintptr_t   flags, cr0reg, cr2reg, cr4reg, new_cr4;
   uintptr_t   drReg;
#ifdef VM_X86_64
   uint64      gs64, fs64;
#endif
   DTR32       gdtreg;
   Selector    cs, gs, fs;
   Descriptor *desc;
   SysenterStateV45 sysenterState;
   VA          vAddr;
   uint16 trReg;   
   uint16 ldtreg;
   unsigned int lint;
   VMCrossPageV45 *crosspage = (VMCrossPageV45*)vm->crosspage[vcpuid];

#if defined(linux) && defined(USE_PERFCTRS_HOSTED)
   PerfCtr_SnapshotFromHost();
#endif
   lint = DisableNMI(vm);
   SAVE_FLAGS(flags); 
   CLEAR_INTERRUPTS();

   vm->currentHostCpu[vcpuid] = HostIF_GetCurrentPCPU();
   
   /*
    * Save CR state (not CR3!)
    */
   GET_CR0(cr0reg);
   GET_CR2(cr2reg);
   GET_CR4(cr4reg);
   
   /* Ensure global pages are flushed */
   new_cr4 = cr4reg & ~CR4_PGE;
   SET_CR4(new_cr4);

   crosspage->switchHostCR4 = new_cr4;
   
   /*
    * Save DR7 since we need to disable debug breakpoints during the
    * world switch code.  Save DR6 in order to accomodate the ICEBP
    * instruction.  All other debug registers are saved lazily by the
    * monitor prior to first use.  NOTE: Since monitor is in legacy
    * mode save all DRs for x86_64 here.
    */

   crosspage->hostDRSaved = 0;

#define SAVE_DR_SHADOW(n) \
                GET_DR##n(drReg); \
                crosspage->hostDR[n] = drReg; \
                crosspage->shadowDR[n] = (uint32) drReg; \
                crosspage->hostDRSaved |= (1 << n);

#ifdef VM_X86_64
   SAVE_DR_SHADOW(0);
   SAVE_DR_SHADOW(1);
   SAVE_DR_SHADOW(2);
   SAVE_DR_SHADOW(3);
#endif // VM_X86_64

   SAVE_DR_SHADOW(6);
   SAVE_DR_SHADOW(7);
   
#undef SAVE_DR_SHADOW
   
   /*
    * Disable debug exceptions during the switch.  Note: If
    * the DR7 GD bit is on, a trap occurred at the GET_DR
    * above, and this process is most likely dead.
    */

   if (UNLIKELY(crosspage->hostDR[7] & DR7_ENABLED)) {
      SET_DR7(DR7_DEFAULT);
      crosspage->shadowDR[7] = DR7_DEFAULT;
   }

#ifdef VM_X86_64
   GET_GDT(crosspage->hostContext64.gdtr64.dtr);
#else
   GET_GDT(gdtreg);
#endif
   GET_LDT(ldtreg); 

   /*
    * GS and FS are saved outside of the SwitchToMonitor() code to 1)
    * minimize the amount of code handled there and 2) prevent us from
    * faulting if they happen to be in the LDT (since the LDT is saved and
    * restored here too) and 3) make sure that if we do fault (because the
    * uncached descriptor for GS or FS's Selector has become invalid) we
    * fault in the host's context rather than the monitor or, worse,
    * never-never land. --Jeremy.
    */
#ifdef VM_X86_64
   gs64 = GET_GS64();
   fs64 = GET_FS64();
#endif
   gs = GET_GS();
   fs = GET_FS();
   
   SysenterReturnToMonitorOld(&sysenterState, crosspage->monUsesSysenter);

   /*
    * Mark our task descriptor as unused, or we won't be
    * able to come back to it (SET_TR set the busy bit)
    */
   
   GET_TR(trReg);
   if (vm_x86_64) {
      crosspage->hostContext64.tr = trReg;
   } else {
      crosspage->hostContext.tr = trReg;
   }

   if (trReg) {
      if (vm_x86_64) {
         vAddr = HOST_KERNEL_LA_2_VA(crosspage->hostContext64.gdtr64.dtr.offset) + trReg;
      } else {
         vAddr = HOST_KERNEL_LA_2_VA(gdtreg.offset) + trReg;
      }
      desc = (Descriptor *)(vAddr);
      if (Desc_Type(desc) == TASK_DESC_BUSY) {
         Desc_SetType(desc, TASK_DESC);
      }
   }

   /*
    * Save the CS on this side
    */
   cs = GET_CS();
   if (vm_x86_64) {
      crosspage->hostContext64.context.cs = cs;
   } else {
      *(uint16*) (&crosspage->hostContext.jump[4]) = cs;
      crosspage->hostContext.task.cs = cs;
   }

   /* See comment for SwitchToMonitor() */
   SwitchToMonitor(crosspage, vm->vmVersion == VME_V45);

   SysenterReturnFromMonitorOld(&sysenterState, crosspage->monUsesSysenter);
   SET_CR0(cr0reg);
   SET_CR2(cr2reg);
   SET_CR4(cr4reg);
   
   SET_LDT(ldtreg);
   
   SET_FS(fs);
   SET_GS(gs);
#ifdef VM_X86_64
   SET_FS64(fs64);
   SET_GS64(gs64);
#endif

#ifdef VM_X86_64
#define CAST_HOST_DR(x) (x)
#else
#define CAST_HOST_DR(x) ((uint32) x)
#endif

#define RESTORE_DR(n) if ((crosspage->hostDRSaved & (1 << n)) && \
                          (crosspage->hostDR[n] != crosspage->shadowDR[n])) { \
                          drReg = CAST_HOST_DR(crosspage->hostDR[n]); \
                          SET_DR##n(drReg); \
                      }

   RESTORE_DR(0);
   RESTORE_DR(1);
   RESTORE_DR(2);
   RESTORE_DR(3);
   RESTORE_DR(6);
   RESTORE_DR(7);

#undef RESTORE_DR
#undef CAST_HOST_DR
   
   ASSERT_NO_INTERRUPTS();

   if (crosspage->isMCIntr()) {
      /*
       * Note we must do the RAISE_INTERRUPT before ever enabling
       * interrupts or bad things have happened (might want to know exactly
       * what bad things btw).
       * Note2: RAISE_INTERRUPT() only takes an constant and hence with switch
       * statement.
       */

#define IRQ_INT(_x) case _x: RAISE_INTERRUPT(_x); break 
#define IRQ_INT2(_x) IRQ_INT(_x); IRQ_INT(_x + 1)
#define IRQ_INT4(_x) IRQ_INT2(_x); IRQ_INT2(_x + 2)
#define IRQ_INT8(_x) IRQ_INT4(_x); IRQ_INT4(_x + 4)
#define IRQ_INT16(_x) IRQ_INT8(_x); IRQ_INT8(_x + 8)
#define IRQ_INT32(_x) IRQ_INT16(_x); IRQ_INT16(_x + 16)

      switch (crosspage->args[0]) {
	 // These are the general IO interrupts
	 // It would be nice to generate this dynamically, but see Note2 above.
#ifdef VM_X86_64
	 IRQ_INT32(0x00);
#else
	 IRQ_INT(0x12);
#endif
	 IRQ_INT32(0x20);
	 IRQ_INT32(0x40);
	 IRQ_INT32(0x60);
	 IRQ_INT32(0x80);
	 IRQ_INT32(0xa0);
	 IRQ_INT32(0xc0);
	 IRQ_INT32(0xe0);

      default: 
	 /*
	  * XXXX nt
	  * running on a 2 processor machine we hit this Panic with int 0xD1 0x61 ...
	  */
	 Warning("Received Unexpected Interrupt: 0x%X in Task_Switch()\n", crosspage->args[0]);
	 Panic("Received Unexpected Interrupt: 0x%X\n", crosspage->args[0]);
      }
   }
   
   vm->currentHostCpu[vcpuid] = INVALID_HOST_CPU;

   RESTORE_FLAGS(flags);
   RestoreNMI(vm, lint);
#if defined(linux) && defined(USE_PERFCTRS_HOSTED)
   PerfCtr_SnapshotFromMonitor();
#endif
}


#ifndef VM_X86_64
/*
 *------------------------------------------------------------------------------
 *
 * Task_InitCrosspage
 *
 *    Initialize the crosspage used to switch to the monitor task. 
 *
 * Results:
 *    0 on success
 *    != 0 on failure
 *
 * Side effects:
 *    None
 *  
 *------------------------------------------------------------------------------
 */

template <class VMCP> static int 
Task_InitCrosspage(VMDriver *vm,          // IN
                   InitBlock *initParams) // IN: Initial params from the VM 
{
   LA baseLinearAddr = HOST_KERNEL_VA_2_LA(0);
   uint32 cr3reg;
   Vcpuid vcpuid;

   ASSERT(sizeof(VMCP) < PAGE_SIZE);
   ASSERT(MODULECALL_CROSS_PAGE_LEN==1);
   for (vcpuid = 0; vcpuid < initParams->numVCPUs;  vcpuid++) {
      /*
       * The physical memory associated to the crosspage has already been locked
       * in MonitorAllocateMemory() --hpreg
       */
      VMCP *p = (VMCP*)HostIF_UserToDriverPtr(vm, initParams->crosspage[vcpuid]);
      if (p == NULL) {
         return 1;
      }
      vm->crosspage[vcpuid] = p;
      GET_CR3(cr3reg);
      p->hostContext.task.cr3 = cr3reg;
      p->hostVA = (VA32)p;
      p->hostContextVA = (uintptr_t)&p->hostContext;
      p->crosspageLinearAddr = baseLinearAddr + (uintptr_t)p;
      {
         DTR32 idtrReg;
         GET_IDT(idtrReg);
         p->monContext.idtr.dtr = idtrReg; /* irrelevant */
      }

      {
         uint16 trReg;
         GET_TR(trReg);
         p->hostContext.tr = trReg;
      }

      p->irqRelocateOffset[0] = IRQ_HOST_INTR1_BASE; 
      p->irqRelocateOffset[1] = IRQ_HOST_INTR2_BASE;

#if defined(WINNT_DDK) && defined(VMX86_DEBUG)
      p->vmPtr = vm;
#endif
   }
   return 0;
}
#endif

int 
Task_InitCrosspage(VMDriver *vm,
                   InitBlock *initParams) // Initial params from the VM
{
   switch (vm->vmVersion) {
      case VME_TOT:
         return Task_InitCrosspage_TOT<VMCrossPageTOT>(vm, initParams, 0);
      case VME_V452:
	 return Task_InitCrosspage_TOT<VMCrossPageV45>(vm, initParams, 0);
      case VME_V45:
         return Task_InitCrosspage_TOT<VMCrossPageV45>(vm, initParams, 1);
#ifndef VM_X86_64
      case VME_V4:
         return Task_InitCrosspage_V4(vm, initParams);
      case VME_GSX251:
      case VME_GSX25:
         return Task_InitCrosspage<VMCrossPageGSX25>(vm, initParams);
      case VME_GSX2:
         return Task_InitCrosspage<VMCrossPageGSX2>(vm, initParams);
      case VME_V321:
         return Task_InitCrosspage<VMCrossPageV321>(vm, initParams);
      case VME_V32:
         return Task_InitCrosspage<VMCrossPageV32>(vm, initParams);
      case VME_V3:
         return Task_InitCrosspage<VMCrossPageV3>(vm, initParams);
      case VME_V2:
         return Task_InitCrosspage<VMCrossPageV2>(vm, initParams);
      case VME_GSX1:
         return Task_InitCrosspage<VMCrossPageGSX1>(vm, initParams);
#endif
      default:
         return 0xDEAD;
   }
}

#ifndef VM_X86_64
/*
 *------------------------------------------------------------------------------
 *
 * Task_Switch --
 *
 *      Switches from the host context into the monitor
 *      context. Think of it as a coroutine switch that changes
 *      not only the registers, but also the address space
 *      and all the hardware state. 
 *
 * Results:
 *      Next module call (or user call for that matter) is
 *      returned. 
 *
 * Side effects:
 *      Jump to the other side. Has no direct effect on the
 *      host-visible state except that it might generate an interrupt. 
 *
 *------------------------------------------------------------------------------
 */

template <class VMCP> static void
Task_Switch(VMCP *crosspage,  // IN
            VMDriver *vm)     // IN
{
   DTR32       gdtreg;
   Descriptor *desc;
   uint32      flags, cr0reg, cr2reg, cr4reg;
   uint32      drreg[8];
   SysenterStateV45 sysenterState;
   Selector    gs, fs;
   VA          vAddr;
   uint16 trReg;   
   uint16 ldtreg;
   unsigned int lint;
  
#if defined(linux) && defined(USE_PERFCOUNTERS)
   PerfCtr_SnapshotFromHost();
#endif
   lint = DisableNMI(vm);
   SAVE_FLAGS(flags); 
   CLEAR_INTERRUPTS();

   /*
    * Save CR state (not CR3!)
    */
   GET_CR0(cr0reg);
   GET_CR2(cr2reg);
   GET_CR4(cr4reg);

   /* Ensure global pages are flushed */
   SET_CR4(cr4reg & ~CR4_PGE);

#define SAVE_DR_SHADOW(n) \
                GET_DR##n(drreg[n]); \
                crosspage->setDR(n, drreg[n]);

   SAVE_DR_SHADOW(0);
   SAVE_DR_SHADOW(0);
   SAVE_DR_SHADOW(1);
   SAVE_DR_SHADOW(2);
   SAVE_DR_SHADOW(3);
   SAVE_DR_SHADOW(6);
   SAVE_DR_SHADOW(7);

#undef SAVE_DR_SHADOW

   /*
    * Disable debug exceptions during the switch.  Note: If
    * the DR7 GD bit is on, a trap occurred at the GET_DR0
    * above, and this process is most likely dead.
    */

   if (UNLIKELY(drreg[7] & DR7_ENABLED)) {
      SET_DR7(DR7_DEFAULT);
      crosspage->setDR(7, DR7_DEFAULT);
   }

   GET_GDT(gdtreg); 
   GET_LDT(ldtreg); 

   /*
    * GS and FS are saved outside of the SwitchToMonitor() code to 1)
    * minimize the amount of code handled there and 2) prevent us from
    * faulting if they happen to be in the LDT (since the LDT is saved and
    * restored here too) and 3) make sure that if we do fault (because the
    * uncached descriptor for GS or FS's Selector has become invalid) we
    * fault in the host's context rather than the monitor or, worse,
    * never-never land. --Jeremy.
    */
   gs = GET_GS();
   fs = GET_FS();

   SysenterReturnToMonitorOld(&sysenterState, crosspage->monUsesSysenter);

   /*
    * Mark our task descriptor as unused, or we won't be
    * able to come back to it (SET_TR set the busy bit)
    * 
    */
   
   GET_TR(trReg);
   crosspage->hostContext.tr = trReg;
   if (trReg) {
     vAddr = HOST_KERNEL_LA_2_VA(gdtreg.offset) + trReg;
     desc = (Descriptor *)(vAddr);
     if (Desc_Type(desc) == TASK_DESC_BUSY) {
       Desc_SetType(desc, TASK_DESC);
     }
   }


   /*
    * Save the CS on this side
    */

   *(uint16*) (&crosspage->hostContext.jump[4]) = GET_CS();
   crosspage->hostContext.task.cs = GET_CS();
   
   /*
    * see comment in taskswitch.h 
    */

   SwitchToMonitor(crosspage, 0);

   SysenterReturnFromMonitorOld(&sysenterState, crosspage->monUsesSysenter);
   SET_CR0(cr0reg);
   SET_CR2(cr2reg);
   SET_CR4(cr4reg);
   
   SET_LDT(ldtreg);  
   
   SET_FS(fs);
   SET_GS(gs);
   
#define RESTORE_DR(n) if (crosspage->isDRModified(n, drreg[n])) \
                              SET_DR##n(drreg[n])
   RESTORE_DR(0);
   RESTORE_DR(1);
   RESTORE_DR(2);
   RESTORE_DR(3);
   RESTORE_DR(6);
   RESTORE_DR(7);

#undef RESTORE_DR

   ASSERT_NO_INTERRUPTS();

   if (crosspage->isMCIntr()) {
      /*
       * Note we must do thje RAISE_INTERRUPT before ever enabling 
       * interrupts or bad things have happened (might want to know exactly
       * what bad things btw). 
       * Note2; RAISE_INTERRUPT() only takes an constant and hence with switch
       * statement. 
       */
#define IRQ_INT(_x) case _x: RAISE_INTERRUPT(_x); break 
#define IRQ_INT2(_x) IRQ_INT(_x); IRQ_INT(_x + 1)
#define IRQ_INT4(_x) IRQ_INT2(_x); IRQ_INT2(_x + 2)
#define IRQ_INT8(_x) IRQ_INT4(_x); IRQ_INT4(_x + 4)
#define IRQ_INT16(_x) IRQ_INT8(_x); IRQ_INT8(_x + 8)
#define IRQ_INT32(_x) IRQ_INT16(_x); IRQ_INT16(_x + 16)
      switch (crosspage->args[0]) {
	 // These are the general IO interrupts
	 // It would be nice to generate this dynamically, but see Note2 above.
	 IRQ_INT(0x12);
	 IRQ_INT32(0x20);
	 IRQ_INT32(0x40);
	 IRQ_INT32(0x60);
	 IRQ_INT32(0x80);
	 IRQ_INT32(0xa0);
	 IRQ_INT32(0xc0);
	 IRQ_INT32(0xe0);

      default: 
	 /*
	  * XXXX nt
	  * running on a 2 processor machine we hit this Panic with int 0xD1 0x61 ...
	  */
	 Warning("Received Unexpected Interrupt: 0x%X in Task_Switch()\n", crosspage->args[0]);
	 Panic("Received Unexpected Interrupt: 0x%X\n", crosspage->args[0]);
      }
   }
   
   RESTORE_FLAGS(flags);
   RestoreNMI(vm, lint);
#if defined(linux) && defined(USE_PERFCOUNTERS)
   PerfCtr_SnapshotFromMonitor();
#endif
      
   return;
}

/*
 *----------------------------------------------------------------------
 *
 * Vmx86_RunVM  --
 *
 *      Let's move RunVM here from Vmx86... linux/signal.h contains
 *      code which does 'unsigned long int qq = -1', and this causes
 *      warnings... And we can make template products static then...
 *
 *----------------------------------------------------------------------
 */
 
/*
 *----------------------------------------------------------------------
 *
 * Vmx86_RunVM_V4  --
 *
 *      Main interaction between the module and the monitor:
 *
 *	   Run the monitor
 *	   Process module calls from the monitor
 *	   Make cross user calls to the main thread
 *	   Return to userlevel to process normal user calls
 *	   and to signal timeout or errors.
 *
 * Results:
 *      Positive: user call number.
 *	-1: error (currently not used).
 *
 * Side effects:
 *      Not really.
 *
 *----------------------------------------------------------------------
 */

template <class VMCP> static int
Vmx86_RunVM_V4(VMDriver *vm, Vcpuid vcpuid)
{
   uint32 retval = MODULECALL_V4_USERRETURN;
   VMCP *crosspage = (VMCP*)vm->crosspage[vcpuid];

   ASSERT(crosspage);

   /*
    * Check if we were interrupted by signal.
    */
   if (crosspage->moduleCallInterrupted) {
      crosspage->moduleCallInterrupted = FALSE;
      goto skipTaskSwitch;
   }
   for (;;) {

      crosspage->retval = retval;

      /*
       * Task_Switch changes the world to the monitor.
       * The monitor is waiting in the BackToHost routine.
       */
      UCTIMESTAMP(crosspage, SWITCHING_TO_MONITOR);
      Task_Switch(crosspage, vm);
      UCTIMESTAMP(crosspage, SWITCHED_TO_MODULE);

skipTaskSwitch:;

      retval = MODULECALL_V4_USERRETURN;

      if (crosspage->userCallType != MODULECALL_USERCALL_NONE) {
         /*
          * This is the main user call path.
          *
          * There are two kinds of user calls.  Normal ones
          * are handled by the calling VCPU thread itself.
          * We just return from here (back to userlevel)
          * in that case.
          *
          * Calls marked userCallCross are handled by the main
          * VMX thread.  In this case, the userCallRequest field
          * indicates to the VMX that this VCPU wants to make
          * a user call.  This field may be consulted by the VMX
          * at any time (specifically when the VMX is awakened by
          * another VCPU), so it must be set after the other
          * user call arguments.  The VMX is responsible for
          * resetting this field and awakening the VCPU when
          * the user call is complete, via the ACK_USER_CALL
          * and COMPLETE_USER_CALL ioctl.  The latter implies
          * the former.
          *
          * When and how to use ACK_USER_CALL and COMPLETE_USER_CALL
          * are at the discretion of the VMX.  In particular,
          * COMPLETE_USER_CALL does not imply that the requested
          * operation has fully completed, only that the VCPU can
          * continue.  See the comment in MonitorLoopHandleUserCall()
          * for use scenarios.
          *
          * See also comment at MonitorLoopUserCallPoll().
          *
          * -- edward
          */

         if (!crosspage->userCallCross) {
            ASSERT(!crosspage->userCallRestart);
            return crosspage->userCallType;
         }

	 if (!crosspage->userCallRestart) {
	    ASSERT(crosspage->userCallRequest == MODULECALL_USERCALL_NONE);
	    crosspage->userCallRequest = crosspage->userCallType;
	    UCTIMESTAMP(crosspage, AWAKENING_VMX);
	    HostIF_UserCall(vm, vcpuid);
	 }

	 UCTIMESTAMP(crosspage, GOING_TO_SLEEP);
         if (HostIF_UserCallWait(vm, vcpuid, USERCALL_TIMEOUT)) {
	    ASSERT(crosspage->userCallRequest == MODULECALL_USERCALL_NONE);
	 } else {
	    retval = MODULECALL_V4_USERTIMEOUT;
	 }
	 UCTIMESTAMP(crosspage, AWAKE);
      }

      switch (crosspage->moduleCallType) {

      case MODULECALL_V4_NONE:
         break;

      case MODULECALL_V4_INTR:    // Already done in task.c
         break;

      case MODULECALL_V4_ISMPNLOCKED : {
         MPN32 mpn = crosspage->args[0];
         retval = HostIF_IsMPNLocked(vm, mpn);
         break;
      }

      case MODULECALL_V4_LOCKPAGE: {
         void *addr = (void *)crosspage->args[0];
	 ASSERT(sizeof retval == sizeof (MPN32));
         retval = Vmx86_LockPage(vm, addr, FALSE);
	 if (PAGE_LOCK_SUCCESS(retval)) {
            if (HostIF_MarkPageDirty(vm, addr)) {
               retval = Vmx86_UnlockPage(vm, addr);
               if (PAGE_LOCK_SUCCESS(retval)) {
                  retval = PAGE_LOCK_TOUCH_FAILED;
               }
            }
         }
         break;
      }

      case MODULECALL_V4_GET_RECYCLED_PAGE:
	 retval = Vmx86_GetRecycledPage(vm);
	 break;

      case MODULECALL_V4_SEMAWAIT:
         retval = HostIF_SemaphoreWait(vm, vcpuid, crosspage->args[0], USERCALL_TIMEOUT);
	 if (retval == MX_WAITINTERRUPTED) {
	    crosspage->moduleCallInterrupted = TRUE;
	    return USERCALL_RESTART;
	 }
         break;
      case MODULECALL_V4_SEMASIGNAL:
         retval = HostIF_SemaphoreSignal(crosspage->args[0]);
	 if (retval == MX_WAITINTERRUPTED) {
	    crosspage->moduleCallInterrupted = TRUE;
	    return USERCALL_RESTART;
	 }
         break;
      case MODULECALL_V4_SEMAFORCEWAKEUP: {
         HostIF_SemaphoreForceWakeup(vm, (Vcpuid) crosspage->args[0]);
         break;
      }
      case MODULECALL_V4_IPI:
         retval = HostIF_IPI(vm, ~0, TRUE);
         break;

      case MODULECALL_V4_UNLOCKPAGE: {
         void *addr = (void *)crosspage->args[0];
         Bool dirty = (Bool)crosspage->args[1];

         if (dirty) {
	    if (HostIF_MarkPageDirty(vm, addr)) {
	       retval = PAGE_UNLOCK_TOUCH_FAILED;
	    } else {
	       retval = Vmx86_UnlockPage(vm, addr);
	    }
         } else {
	    retval = Vmx86_UnlockPage(vm, addr);
	 }
         break;
      }

      case MODULECALL_V4_RELEASE_ANON_PAGE: {
         MPN32 mpn = (MPN32)crosspage->args[0];
	 retval = Vmx86_ReleaseAnonPage(vm, mpn);
         break;
      }

      default:
         Warning("ModuleCall %d not supported\n", crosspage->moduleCallType);
      }

      /*
       * Give other threads and processes a chance.
       *
       * This is important when we are not preemptable while
       * in the driver and the monitor (e.g., Linux).
       */

      HostIF_Reschedule();

   }

   NOT_REACHED();
}


/*
 *----------------------------------------------------------------------
 *
 * Vmx86_RunVM  --
 *
 *      Return to the monitor.
 *      Main interaction between the module and the monitor.
 *      Leaves the loop only to go back to user mode.
 *      Every interation of the loop switches back to the monitor
 *
 * Results:
 *      Returns the MODULECALL that forced it to leave the loop.
 *      back to the IOCTL handler of the module device driver
 *
 * Side effects:
 *      Dispatches the messages, everything can change
 *
 *----------------------------------------------------------------------
 */
template <class VMCP> static int
Vmx86_RunVM(VMCP *crosspage, VMDriver *vm)
{
   ModuleCallTypeV3 op;
   uint32 retval;

   ASSERT(vm && crosspage);

   retval = MODULECALL_V3_USERRETURN;

   while(1) {

      crosspage->retval = retval;

      /*
       * Task_Switch changes the world to the monitor.
       * The monitor is waiting in the BackToHost routine
       */
      Task_Switch(crosspage, vm);

      op = crosspage->moduleCallType;

      retval = MODULECALL_USER_START;

      if (op >  MODULECALL_USER_START&& op < MODULECALL_USER_END) {
         return op;
      }


      switch (op) {
      case MODULECALL_V3_NONE :
         break;


      case MODULECALL_V3_INTR : {
	 /*
	  * Note that as of SMP support, irq is the actual Interrupt number
	  * rather than the IRQ, b/c the IO APIC redirects IRQs every which
	  * way. --Jeremy.
	  */
#ifdef notdef
	 if (crosspage->args[1] == MODULECALL_USERCALL_NONE) {
	    /* nothing to be done if MODULECALL_V3_USERRETURN */
	    retval = MODULECALL_V3_USERRETURN;
	    break;
	 }
#endif
	 return crosspage->args[1];
      }
      case MODULECALL_V3_ISMPNLOCKED : {
         MPN mpn = crosspage->args[0];
         retval = HostIF_IsMPNLocked(vm, mpn);
         break;
      }

      case MODULECALL_V3_LOCKPAGE: {
         char dummy;
         void *addr = (void *)crosspage->args[0];
	 ASSERT(sizeof retval == sizeof (MPN));
         retval = Vmx86_LockPage(vm, addr, FALSE);
	 if (PAGE_LOCK_SUCCESS(retval)) {
            if (HostIF_CopyFromUser(&dummy, addr, 1) == 0) {
               HostIF_CopyToUser(addr, &dummy, 1);
            } else {
               retval = Vmx86_UnlockPage(vm, addr);
	       if (PAGE_LOCK_SUCCESS(retval)) {
		  retval = PAGE_LOCK_TOUCH_FAILED;
	       }
            }
         }
         break;
      }

      case MODULECALL_V3_UNLOCKPAGE: {
         char dummy;
         void *addr = (void *)crosspage->args[0];
         Bool dirty = (Bool)crosspage->args[1];

         if (dirty) {
            if (HostIF_CopyFromUser(&dummy, addr, 1) != 0 ||
                HostIF_CopyToUser(addr, &dummy, 1) != 0) {
	       retval = PAGE_UNLOCK_TOUCH_FAILED;
            } else {
	       retval = Vmx86_UnlockPage(vm, addr);
	    }
         } else {
            retval = Vmx86_UnlockPage(vm, addr);
         }
         break;
      }

      default:
         Panic("ModuleCallLoop %d not supported \n",op);
      }
   }

   ASSERT(op >= MODULECALL_USER_START && op < MODULECALL_USER_END);
   return op;
}
#endif

int Vmx86_RunVM(VMDriver *vm, Vcpuid vcpuid) {
#ifndef VM_X86_64
   void* crosspage = vm->crosspage[vcpuid];
#endif
   switch (vm->vmVersion) {
#ifndef VM_X86_64
      case VME_V2:
         return Vmx86_RunVM((VMCrossPageV2*)crosspage, vm);
      case VME_GSX1:
         return Vmx86_RunVM((VMCrossPageGSX1*)crosspage, vm);
      case VME_V3:
         return Vmx86_RunVM((VMCrossPageV3*)crosspage, vm);
      case VME_V321:
      	 return Vmx86_RunVM((VMCrossPageV321*)crosspage, vm);
      case VME_V32:
         return Vmx86_RunVM((VMCrossPageV32*)crosspage, vm);
      case VME_GSX2:
         return Vmx86_RunVM((VMCrossPageGSX2*)crosspage, vm);
      case VME_GSX251:
      case VME_GSX25:
         return Vmx86_RunVM((VMCrossPageGSX25*)crosspage, vm);
      case VME_V4:
         return Vmx86_RunVM_V4<VMCrossPageV4>(vm, vcpuid);
#endif
      case VME_V45:
         return Vmx86_RunVM_V45(vm, vcpuid);
      case VME_V452:
	 return Vmx86_RunVM_V45(vm, vcpuid);
      case VME_TOT:
         return Vmx86_RunVM_TOT(vm, vcpuid);
      default:
         return -EINVAL;
   }
}

/*
 *----------------------------------------------------------------------
 *
 * CheckPadding --
 *
 *      check for expected padding --
 *      this check currently fails on the egcs compiler
 *
 * Results:
 *
 *      TRUE if the check succeeds -- module will be loaded
 *
 *
 *
 * Side effects:
 *      output to kernel log on error
 *      Structure size check may fail at build time, not at load time.
 *
 *----------------------------------------------------------------------
 */

#define VM_CHECKSIZE(structure, expected)       \
   if (sizeof(structure) != expected) {         \
      asm volatile(".print \"Unexpected ABI change: Structure " #structure " should take %c0 bytes, but took %c1!\"; .abort" : : "i"(expected), "i"(sizeof(structure))); \
   }

#define VM_LACHECK(structure, expected)							\
   if (offsetof(structure, crosspageLinearAddr) != expected) {				\
      asm volatile(".print \"Unexpected ABI change: Structure " #structure " should have crosspageLinearAddr at offset %c0, but has it at %c1!\"; .abort" : : \
		   "i"(expected), "i"(offsetof(structure, crosspageLinearAddr)));	\
   }

int
Task_CheckPadding(void)
{
   DTRWords32 dtr;
   uint16 *x;
   int bad = 0;

   memset(&dtr, 0, sizeof dtr);
   dtr.dtr.limit = 0x1111;
   dtr.dtr.offset = 0x22223333;

   x = (uint16*)&dtr;

   if (x[0] == 0x1111 && x[1] == 0x3333 && x[2] == 0x2222) {
   } else {
      Warning("DTR padding\n");
      bad = 1;
   }

   VM_CHECKSIZE(VMCrossPageTOT,   MODULECALL_CROSSPAGE_SIZE_TOT);
   VM_LACHECK(VMCrossPageTOT,	  MODULECALL_CROSSPAGE_LAOFFSET_TOT);
   VM_CHECKSIZE(VMCrossPageV45,   MODULECALL_CROSSPAGE_SIZE_V45);
   VM_LACHECK(VMCrossPageV45,     MODULECALL_CROSSPAGE_LAOFFSET_V45);
   VM_CHECKSIZE(VMCrossPageV4,    MODULECALL_CROSSPAGE_SIZE_V4);
   VM_LACHECK(VMCrossPageV4,      MODULECALL_CROSSPAGE_LAOFFSET_V4);
   VM_CHECKSIZE(VMCrossPageGSX25, MODULECALL_CROSSPAGE_SIZE_GSX25);
   VM_LACHECK(VMCrossPageGSX25,   MODULECALL_CROSSPAGE_LAOFFSET_GSX25);
   VM_CHECKSIZE(VMCrossPageGSX2,  MODULECALL_CROSSPAGE_SIZE_GSX2);
   VM_LACHECK(VMCrossPageGSX2,    MODULECALL_CROSSPAGE_LAOFFSET_GSX2);
   VM_CHECKSIZE(VMCrossPageV321,  MODULECALL_CROSSPAGE_SIZE_V321);
   VM_LACHECK(VMCrossPageV321,    MODULECALL_CROSSPAGE_LAOFFSET_V321);
   VM_CHECKSIZE(VMCrossPageV32,   MODULECALL_CROSSPAGE_SIZE_V32);
   VM_LACHECK(VMCrossPageV32,     MODULECALL_CROSSPAGE_LAOFFSET_V32);
   VM_CHECKSIZE(VMCrossPageV3,    MODULECALL_CROSSPAGE_SIZE_V3);
   VM_LACHECK(VMCrossPageV3,      MODULECALL_CROSSPAGE_LAOFFSET_V3);
   VM_CHECKSIZE(VMCrossPageV2,    MODULECALL_CROSSPAGE_SIZE_V3);
   VM_LACHECK(VMCrossPageV2,      MODULECALL_CROSSPAGE_LAOFFSET_V3);
   VM_CHECKSIZE(VMCrossPageGSX1,  MODULECALL_CROSSPAGE_SIZE_V3);
   VM_LACHECK(VMCrossPageGSX1,    MODULECALL_CROSSPAGE_LAOFFSET_V3);

   if (!bad) {
      return TRUE;
   }

   printk("/dev/vmmon: Cannot load module. Use standard gcc compiler\n");
   return FALSE;
}
